import tkinter as tk
import tkinter.messagebox as tkmessagebox
import tkinter.filedialog as tkfiledialog
# from time import *
from os import mkdir
from os.path import isfile, isdir
from operator import itemgetter
from Marker_Finder_prepare import gb_to_fasta, gRNA_candidates_lr, make_blast_database
from first_filter import first_filter_db_lr, first_filter_sbjct_lr
from second_filter import second_filter_lr
from PCRFragmentDesigner import PCRFD_Win


def make_dir(filedir):
    if not isdir(filedir):
        mkdir(filedir)


def isoverlap(x, y):
    return y[0] <= x[1] and x[0] <= y[1]


def data_read(infname):
    InputFile = open(infname)
    lst = []
    for line in InputFile:
        a = line.strip()
        a = a.split()
        lst.append((a[0], int(a[1]), float(a[2])))
    InputFile.close()
    return lst


def pair_score(g1_score, g2_score, gap):
    return (g1_score+g2_score)/2


class MF_Win(tk.Frame):
    def __init__(self):
        tk.Frame.__init__(self)
        self.master.title("BioSeqVisualizationDesigner")
        self.master.geometry("1280x960")  # 设置主窗口大小
        self.grid()

        # Target Genome
        self._TargetLabel = tk.Label(self, text="Target Genome Name",
                                     font=('Calibri', 18, 'bold'))
        self._TargetLabel.grid(row=0, column=0)
        self._TargetNameVar = tk.StringVar()
        self._TargetFilenameVar = tk.StringVar()
        self._TargetFileformatVar = tk.StringVar()
        self._TargetNameEntry = tk.Entry(
            self, textvariable=self._TargetNameVar, width=2,
            font=('Calibri', 18, 'bold'))
        self._TargetNameEntry.grid(row=0, column=1)
        self._TargetFileButton = tk.Button(self, text="Choose Target File",
                                           font=('Calibri', 18, 'bold'),
                                           command=self.choose_target_file)
        self._TargetFileButton.grid(row=0, column=2)

        # Subject Genome
        self._SubjectLabel = tk.Label(self, text="Subject Genome Name",
                                      font=('Calibri', 18, 'bold'))
        self._SubjectLabel.grid(row=1, column=0)
        self._SubjectNameVar = tk.StringVar()
        self._SubjectFilenameVar = tk.StringVar()
        self._SubjectFileformatVar = tk.StringVar()
        self._SubjectNameEntry = tk.Entry(
            self, textvariable=self._SubjectNameVar, width=2,
            font=('Calibri', 18, 'bold'))
        self._SubjectNameEntry.grid(row=1, column=1)
        self._SubjectFileButton = tk.Button(self, text="Choose Subject File",
                                            font=('Calibri', 18, 'bold'),
                                            command=self.choose_subject_file)
        self._SubjectFileButton.grid(row=1, column=2)

        # Genome Site Set
        self._SiteRangeLabel = tk.Label(self, text="Genome Position Range",
                                        font=('Calibri', 18, 'bold'))
        self._SiteRangeLabel.grid(row=2, column=0)
        self._StartPositionVar = tk.IntVar()
        self._EndPositionVar = tk.IntVar()
        self._StartPositionEntry = tk.Entry(
            self, textvariable=self._StartPositionVar, width=2,
            font=('Calibri', 18, 'bold'))
        self._StartPositionEntry.grid(row=2, column=1)
        self._EndPositionEntry = tk.Entry(
            self, textvariable=self._EndPositionVar, width=2,
            font=('Calibri', 18, 'bold'))
        self._EndPositionEntry.grid(row=2, column=2)

        # Spacer Range Set
        self._SpacerLabel = tk.Label(self, text="Pair Spacer Range",
                                     font=('Calibri', 18, 'bold'))
        self._SpacerLabel.grid(row=3, column=0)
        self._MinLengthVar = tk.IntVar()
        self._MaxLengthVar = tk.IntVar()
        self._MinLengthEntry = tk.Entry(
            self, textvariable=self._MinLengthVar, width=2,
            font=('Calibri', 18, 'bold'))
        self._MinLengthEntry.grid(row=3, column=1)
        self._MaxLengthEntry = tk.Entry(
            self, textvariable=self._MaxLengthVar, width=2,
            font=('Calibri', 18, 'bold'))
        self._MaxLengthEntry.grid(row=3, column=2)

        self._Orientation_Var = tk.IntVar()
        self._Orientation_Var.set(1)
        self._Orientation_Label = tk.Label(self, text="Orientation",
                                           font=('Calibri', 18, 'bold'))
        self._Orientation_Label.grid(row=0, column=3)
        self._orientation = tk.Radiobutton(self, variable=self._Orientation_Var,
                                           value=1,
                                           text='Orientation eL+eR:CCNN20...N20NGG')
        self._orientation.grid(row=0, column=4)
        self._orientation = tk.Radiobutton(self, variable=self._Orientation_Var,
                                           value=2,
                                           text='Orientation iL+iR:N20NCC...GGNN20')
        self._orientation.grid(row=1, column=4)
        self._orientation = tk.Radiobutton(self, variable=self._Orientation_Var,
                                           value=3,
                                           text='Orientation eL+iR:CCNN20...CCNN20')
        self._orientation.grid(row=2, column=4)
        self._orientation = tk.Radiobutton(self, variable=self._Orientation_Var,
                                           value=4,
                                           text='Orientation iL+eR:N20NCC...N20NCC')
        self._orientation.grid(row=3, column=4)

        self._CandidateSequence = tk.Text(self, height=30, width=160)
        self._CandidateSequence.grid(row=6, column=0, columnspan=6)

        self._finderbutton = tk.Button(self, text="Finder",
                                       font=('Calibri', 18, 'bold'),
                                       command=self.spacer_finder_prepare)
        self._finderbutton.grid(row=5, column=1)

        # self._sort_by_gap_button = tk.Button(self, text = 'Sort by gap',
        #                              font=('Calibri', 24, 'bold'),
        #                              command = lambda : self._sort_by_gap(listCCN,listNGG,listCCN_B,listNGG_B))
        # self._sort_by_gap_button.grid(row=2,column=2)
        self._sort_by_pos_button = tk.Button(self, text="Sort by pos",
                                             font=('Calibri', 18, 'bold'),
                                             command=self._sort_by_pos)
        self._sort_by_pos_button.grid(row=5, column=2)
        self._savebutton = tk.Button(self, text='Save',
                                     font=('Calibri', 18, 'bold'),
                                     command=self._save)
        self._savebutton.grid(row=5, column=3)
        self._PCR_design_button = tk.Button(self, text='PCR fragment Design',
                                            font=('Calibri', 18, 'bold'),
                                            command=self.PCR_fragment_design)
        self._PCR_design_button.grid(row=5, column=4)

    def choose_file(self):
        filename = tkfiledialog.askopenfilename(parent=self, filetypes=(("FASTA files", "*.fasta;*.fsa;*.txt"),
                                                                        ("GenBank files",
                                                                         "*.gb;*.gbk;*.gbff"),
                                                                        ("All files", "*.*")))
        if filename:
            try:
                formatflag = filename.split('.')[1]
                if formatflag in ['fsa', 'fasta', 'txt']:
                    fileformat = 'fasta'
                    return filename, fileformat
                elif formatflag in ['gb', 'gbk', 'gbff']:
                    fileformat = 'gb'
                    return filename, fileformat
                else:
                    raise OSError
            except:
                tkmessagebox.showerror(
                    "Can't Find Gene File (fasta or GenBank): {}".format(filename))
                return '', ''

    def choose_target_file(self):
        filename, fileformat = self.choose_file()
        self._TargetFilenameVar.set(filename)
        self._TargetFileformatVar.set(fileformat)

    def choose_subject_file(self):
        filename, fileformat = self.choose_file()
        self._SubjectFilenameVar.set(filename)
        self._SubjectFileformatVar.set(fileformat)

    def spacer_finder_prepare(self, flag=0):
        filedir = '{}_{}'.format(
            self._TargetNameVar.get(), self._SubjectNameVar.get())
        make_dir(filedir)
        if not (isfile(filedir+'\choose_CCN_gRNA.txt') and isfile(filedir+'\choose_NGG_gRNA.txt')):
            self.spec_test(flag)
        CCN = data_read(filedir+'\choose_CCN_gRNA.txt')
        NGG = data_read(filedir+'\choose_NGG_gRNA.txt')
        orientation_flag = self._Orientation_Var.get()
        if orientation_flag == 1:
            self.spacer_finder(CCN, NGG, 'CCN_NGG', flag)
        elif orientation_flag == 2:
            self.spacer_finder(NGG, CCN, 'NGG_CCN', flag)
        elif orientation_flag == 3:
            self.spacer_finder(CCN, CCN, 'CCN_CCN', flag)
        elif orientation_flag == 4:
            self.spacer_finder(NGG, NGG, 'NGG_NGG', flag)

    def spec_test(self, flag):
        # Search for gRNA Candidates
        if self._SubjectFileformatVar.get() == 'gb':
            self._SubjectFilenameVar.set(
                gb_to_fasta(self._SubjectNameVar.get(), self._SubjectFilenameVar.get()))
        if self._TargetFileformatVar.get() == 'gb':
            self._TargetFilenameVar.set(
                gb_to_fasta(self._TargetNameVar.get(), self._TargetFileformatVar.get()))
        make_dir('db')
        target_filename = self._TargetFilenameVar.get()
        database_name = 'db\\'+self._SubjectNameVar.get()
        make_blast_database(database_name, self._SubjectFilenameVar.get())
        filedir = '{}_{}'.format(
            self._TargetNameVar.get(), self._SubjectNameVar.get())
        make_dir(filedir)
        filename_str = filedir+'\{}_all.fasta'
        gRNA_candidates_lr(
            self._TargetFilenameVar.get(), 'fasta', filename_str)

        # Specificity test for each candidate
        # first_filter
        out_filename_str = filedir+'\{}_gRNA_first_filter_by_self.fasta'
        first_filter_sbjct_lr(filename_str, out_filename_str, target_filename)
        first_filter_out_filename_str = filedir+'\{}_gRNA_first_filter.fasta'
        first_filter_db_lr(
            out_filename_str, first_filter_out_filename_str, database_name)
        # second_filter
        subject_filename = self._SubjectFilenameVar.get()
        second_filter_lr(
            first_filter_out_filename_str, target_filename, subject_filename, database_name)

    def spacer_finder(self, CCN, NGG, gRNA_str, flag):
        min_len = self._MinLengthVar.get()
        if min_len == 0:
            min_len = 0
        max_len = self._MaxLengthVar.get()
        if max_len == 0:
            max_len = 100
        self._clear()
        nearestCCN = [0 for i in range(len(CCN))]
        CCNnum = 0
        NGGnum = 0
        k = 0
        for NGGgRNA in NGG:
            k += 1
            NGGnum += 1
            while CCNnum < len(CCN) and CCN[CCNnum][1] + 20 + max_len < NGGgRNA[1]:
                CCNnum += 1
            while CCNnum < len(CCN) and CCN[CCNnum][1] + 20 + min_len - 1 < NGGgRNA[1]:
                nearestCCN[CCNnum] = NGGnum
                CCNnum += 1
        pair_list = []
        for i in range(len(CCN)):
            if nearestCCN[i] != 0:
                j = nearestCCN[i]-1
                while j < len(NGG):
                    g1 = CCN[i]
                    g2 = NGG[j]
                    gap = (g2[1] - g1[1]-20)
                    if gap <= max_len:
                        score = pair_score(g1[2], g2[2], gap)
                        pair_list.append((g1, g2, gap, score))
                    elif gap > max_len:
                        break
                    j += 1
        candidate_num = len(pair_list)
        self._CandidateSequence.insert(tk.END, "Sequence gRNA_pair\n")
        self._CandidateSequence.insert(
            tk.END, "The number of candidate sequences:{}\n".format(candidate_num))
        k = 0
        filedir = '{}_{}'.format(
            self._TargetNameVar.get(), self._SubjectNameVar.get())
        if flag == 2:
            sort_lst = sorted(pair_list, key=lambda x: x[0][1])
            outfilename = '{}\{}_sort_by_{}.txt'.format(
                filedir, gRNA_str, 'pos')
        elif flag == 1:
            sort_lst = sorted(pair_list, key=itemgetter(2))
            outfilename = '{}\{}_sort_by_{}.txt'.format(
                filedir, gRNA_str, 'gap')
        else:
            sort_lst = sorted(pair_list, key=itemgetter(3), reverse=True)
            outfilename = '{}\{}_sort_by_{}.txt'.format(
                filedir, gRNA_str, 'score')
        OutFile = open(outfilename, 'w')
        for pair in sort_lst:
            k += 1
            self._CandidateSequence.insert(tk.END, "G1#{:>10d}..G2#{:>10d} : {}{:.^40d}{} {:.2f}\n".format(
                pair[0][1], pair[1][1], pair[0][0], pair[2], pair[1][0], pair[3]))
            OutFile.write("G1#{:>10d}..G2#{:>10d} : {}{:.^40d}{} {:.2f}\n".format(
                pair[0][1], pair[1][1], pair[0][0], pair[2], pair[1][0], pair[3]))
        OutFile.close()
        # print('the whole time: {}'.format(time()-t00))

    # def _sort_by_gap(self,listCCN,listNGG,listCCN_B,listNGG_B):
    #     try:
    #         self._clear()
    #         self.spacer_finder_prepare(1)
    #     except:
    #         tkmessagebox.showerror(title="Sort Command", parent = self,
    #                               message = "Error: No Candidate Sequences")
    #     return
    def _sort_by_pos(self):
        try:
            self._clear()
            self.spacer_finder_prepare(2)
        except:
            tkmessagebox.showerror(title="Sort Command", parent=self,
                                   message="Error: No Candidate Sequences")
        return

    def PCR_fragment_design(self):
        filedir = '{}_{}'.format(
            self._TargetNameVar.get(), self._SubjectNameVar.get())
        filename = filedir+'\CCN_NGG_sort_by_pos.txt'
        if isfile(filename):
            PCRFD_Win(self, filename).mainloop()
        else:
            tkmessagebox.showerror(title="gRNA pair", parent=self,
                                   message="Error: Can't find gRNA pair file {}".format(filename))

    def _save(self):
        data = self._CandidateSequence.get('0.0', tk.END)
        filename = tkfiledialog.asksaveasfilename(parent=self, title='Save as...',
                                                  filetypes=[('Text', '*.txt')])
        try:
            if not filename.endswith('.txt'):
                filename += '.txt'
            with open(filename, 'w') as w:
                w.write(data)
        except:
            tkmessagebox.showerror(title="Write File", parent=self,
                                   message="Error: Can't open: {}".format(filename))

    def _clear(self):
        self._CandidateSequence.delete(0.0, tk.END)

MF_Win().mainloop()
