from re import compile
from os import system
from os.path import isfile, dirname, basename
# from os import mkdir
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
from time import time


def gb_to_fasta(targetname, target_filename):
    fastafilename = dirname(target_filename)+'/'+targetname+'.fasta'
    SeqIO.convert(target_filename,'gb',fastafilename,'fasta')
    return fastafilename

def gRNA_candidate_iteration(dna_record, gRNA_flag):
    dna = str(dna_record.seq).lower()
    if gRNA_flag == 'CCN':  # CCNN20
        regexp = compile("(?<=cc).(?=.{20})")
        for mat in regexp.finditer(dna):
            yield SeqRecord(Seq(dna[mat.start()+1:mat.start()+20+1]), 
                id='{}|{}'.format(dna_record.id,mat.start()+2),description='')
    elif gRNA_flag == 'NGG':  # N20NGG
        regexp = compile("(?<=.{20}).(?=gg)")
        for mat in regexp.finditer(dna):
            yield SeqRecord(Seq(dna[mat.start()-20:mat.start()]), 
                id='{}|{}'.format(dna_record.id,mat.start()-20+1),description='')

def gRNA_candidates(dna_record,gRNA_flag,out_filename):
    if isfile(out_filename):
        return
    gRNA_iter = gRNA_candidate_iteration(dna_record,gRNA_flag)
    SeqIO.write(gRNA_iter, out_filename, 'fasta')
    del gRNA_iter

def gRNA_candidates_lr(target_filename,target_format,filename_str):
    dna_record = SeqIO.read(target_filename,target_format)
    for gRNA_flag in ['CCN','NGG']:
        gRNA_candidates(dna_record,gRNA_flag,filename_str.format(gRNA_flag))

def file_combines_iteration(file_list,file_format):
    for file in file_list:
        for record in SeqIO.parse(file,file_format):
            yield record

def subject_combines(subject_file_list,in_format,out_filename, out_format):
    SeqIO.write(file_combines_iteration(subject_file_list,in_format),out_filename,out_format)

def make_blast_database(database_name,subject_filename,max_file_sz='1GB'):
    record_dict = {}
    for record in SeqIO.parse(subject_filename,'fasta'):
        if record.id in record_dict:
            continue
        record_dict.setdefault(record.id,record)
    subject_unique_filename = 'uni_{}'.format(subject_filename)
    SeqIO.write(record_dict.values(),subject_unique_filename,'fasta')
    system('''makeblastdb.exe -in {} -dbtype nucl -out {} -max_file_sz {}'''.format(subject_unique_filename,database_name,max_file_sz))

if __name__=='main':
    t0 = time()
    target_filename = 'ex_target.gb'
    subject_filename = 'ex_db.gb'
    target_filename = gb_to_fasta('ex_target', target_filename)
    subject_filename = gb_to_fasta('ex_db', subject_filename)
    make_blast_database('db\ex_db',subject_filename)
    make_blast_database('db\ex_target',target_filename)
    filename_str = '{}_all.fasta'
    gRNA_candidates_lr(target_filename,'fasta',filename_str)
    print('Prepare ends {}'.format(time()-t0))
