import tkinter as tk
import tkinter.messagebox as tkmessagebox
import tkinter.filedialog as tkfiledialog
from re import finditer
from os import mkdir
from os.path import isdir,dirname

def make_dir(filedir):
    if not isdir(filedir):
        mkdir(filedir)

def isoverlap(x,y):
            return y[0]<=x[1] and x[0]<=y[1]

class PCRFD_Win(tk.Toplevel):
    def __init__(self,parent,gRNApair_filename):
        tk.Toplevel.__init__(self,master=parent)
        self.title("PCR_Fragment_Designer")
        self.geometry("1280x960") # 设置主窗口大小
        self.grid()
        if parent is None:
            self._min_gap = 19
            self._max_gap = 23
        else:
            self._min_gap = self.master._MinLengthVar.get()
            self._max_gap = self.master._MaxLengthVar.get()
        self._Pair_Filename = tk.StringVar()
        self._Pair_Filename.set(gRNApair_filename)
        self._Cluster_Var = tk.IntVar()
        self._Cluster_Var.set(1)
        self._Cluster_Method_Label = tk.Label(self, text = "Cluster Method",
                                     font=('Calibri', 18, 'bold'))
        self._Cluster_Method_Label.grid(row = 0, column = 0)
        self._cluster_finder_flag = tk.Radiobutton(self,variable = self._Cluster_Var,
                                           value=1,
                                           text = 'Fix Fragment Number')
        self._cluster_finder_flag.grid(row = 1, column=0)
        self._cluster_finder_flag = tk.Radiobutton(self,variable = self._Cluster_Var,
                                           value=2,
                                           text = 'Fix Fragment Length')
        self._cluster_finder_flag.grid(row = 2, column=0)

        # Range Set
        self._SearchRangeLabel = tk.Label(self, text = "Search Range (Num/Length)",
                                     font=('Calibri', 18, 'bold'))
        self._SearchRangeLabel.grid(row = 0, column = 1,columnspan=2)
        self._MaxVar = tk.IntVar()
        self._MaxEntry = tk.Entry(
            self, textvariable = self._MaxVar,width = 2,
            font=('Calibri', 18, 'bold'))
        self._MaxEntry.grid(row = 1, column = 2,sticky='W')

        self._CandidatePCRFragment = tk.Text(self,height=30, width=160)
        self._CandidatePCRFragment.grid(row=4,column=0,columnspan=6)

        self._cluster_finder_Button = tk.Button(self, text='PCR Design',font=('Calibri',18,'bold'),command=self.cluster_finder)
        self._cluster_finder_Button.grid(row=3,column=1)

    def cluster_finder(self):
        if self._Cluster_Var.get() == 1:
            self.fix_number_cluster()
        else:
            self.fix_length_cluster()

    def fix_number_cluster(self):
        self._clean()
        filename = self._Pair_Filename.get()
        InputFile = open(filename)
        most_number = self._MaxVar.get()
        s = []
        for line in InputFile:
            matlist = list(finditer('(?<=\W)\d+(?=\W)',line.rstrip('.0123456789\n')))
            if len(matlist) == 3:
                start_pos = int(matlist[0].group())
                gap = int(matlist[2].group())
                end_pos = start_pos + gap + 43-1
                s.append((start_pos-3,end_pos,line))
        InputFile.close()
        v_num = len(s)
        group_num = 100
        min_group_list = [[] for i in range(group_num)]
        min_group_subject_len_list = [1000000 for i in range(group_num)]
        k = 0
        while k < v_num:
            group = []
            start = s[k][0]
            current = k
            while current< v_num and len(group) < most_number:
                group_len = s[current][1] - start + 1
                group.append(current)
                next = current+1
                while next<v_num and isoverlap(s[current],s[next]):
                    next += 1
                current = next
            k += 1
            if len(group)<most_number:
                continue
            i = group_num-1
            if len(group)==len(min_group_list[i]) and group_len<min_group_subject_len_list[i] or len(group)>len(min_group_list[i]):
                while i > 0 and \
                        (len(group)>len(min_group_list[i-1]) or
                                             len(group)== len(min_group_list[i-1]) and group_len<min_group_subject_len_list[i-1]):
                    min_group_list[i] = min_group_list[i-1]
                    min_group_subject_len_list[i] = min_group_subject_len_list[i-1]
                    i -= 1
                min_group_list[i] = group
                min_group_subject_len_list[i] = group_len
        make_dir(dirname(filename)+'/Fix_Num_Test')
        final_lst = []
        num_lst = []
        for i in range(group_num):
            if not min_group_list[i]: continue
            score_lst = []
            lst = []
            num_gap_dict = {ii:0 for ii in range(self._min_gap,self._max_gap+1)}
            score_sum = 0
            OutFile = open('{}\Fix_Num_Test\max_group_{}.txt'.format(dirname(filename),i),'w')
            for j in min_group_list[i]:
                line = s[j][2]
                OutFile.write(line)
                matlist = list(finditer('(?<=\W)\d+(?=\W)',line))
                if len(matlist)==5:
                    start_pos = int(matlist[0].group())
                    gap = int(matlist[2].group())
                    end_pos = start_pos + gap + 43-1
                    score = int(matlist[3].group())+int(matlist[4].group())/100
                    score_sum += score
                    score_lst.append(score)
                    num_gap_dict[gap] += 1
                    lst.append((start_pos-3,end_pos))
            OutFile.close()
            length = lst[-1][1]-lst[0][0]+1
            num_lst.append(len(lst))
            final_lst.append('{:>2d}\t{:>9d}\t{:>9d}\t{:>4d}\t{:>2d}\t{}\t{:>.2f}\t{}\n'.format(i,lst[0][0],lst[-1][1],length,
                                                                                          len(score_lst),'\t'.join((str(num_gap_dict[i]) for i in range(self._min_gap,self._max_gap+1))),
                                                                                          score_sum/len(score_lst),score_lst))
        with open(dirname(filename)+'/Fix_Num_Test/result.txt','w') as OutFile:
            OutFile.write('{:>2s}\t{:>9s}\t{:>9s}\t{:>4s}\t{:>2s}\tgap{}\t{}\t{}\n'.format('#file','start_pos','end_pos','length','#pair','\tgap'.join((str(i) for i in range(self._min_gap,self._max_gap+1))),'score','score_lst'))
            self._CandidatePCRFragment.insert(tk.END,'{:>2s}\t{:>9s}\t{:>9s}\t{:>4s}\t{:>2s}\tgap{}\t{}\t{}\n'.format('#file','start_pos','end_pos','length','#pair','\tgap'.join((str(i) for i in range(self._min_gap,self._max_gap+1))),'score','score_lst'))
            for i in sorted(range(len(final_lst)),key= lambda x:num_lst[x],reverse=True):
                OutFile.write(final_lst[i])
                self._CandidatePCRFragment.insert(tk.END,final_lst[i])
        self._Choose_Cluster_Label = tk.Label(self,text='choose cluster number(seperate by blank)',font=('Calibri', 18, 'bold'))
        self._Choose_Cluster_Label.grid(row = 0, column = 3,columnspan=3)
        self._choose_cluster_number = tk.StringVar()
        self._ChooseClusterNumberEntry = tk.Entry(
            self, textvariable = self._choose_cluster_number,
            font=('Calibri', 18, 'bold'))
        self._ChooseClusterNumberEntry.grid(row = 1, column = 4,columnspan=2)
        self._cluster_finder_Button = tk.Button(self, text='Oligo Generator',font=('Calibri',18,'bold'),command=lambda : self.oligo_generator('Fix_Num_Test'))
        self._cluster_finder_Button.grid(row=3,column=2)
        return

    def fix_length_cluster(self):
        self._clean()
        filename = self._Pair_Filename.get()
        max_len = self._MaxVar.get()
        InputFile = open(filename)
        s = []
        for line in InputFile:
            matlist = list(finditer('(?<=\W)\d+(?=\W)',line.rstrip('.0123456789\n')))
            if len(matlist) == 3:
                start_pos = int(matlist[0].group())
                gap = int(matlist[2].group())
                end_pos = start_pos + gap + 43-1
                s.append((start_pos-3,end_pos,line))
        InputFile.close()
        v_num = len(s)
        group_num = 100
        max_group_list = [[] for i in range(group_num)]
        k = 0
        while k < v_num:
            group = []
            start = s[k][0]
            current = k
            while current< v_num and s[current][1] - start + 1 <= max_len:
                group.append(current)
                next = current+1
                while next<v_num and isoverlap(s[current],s[next]):
                    next += 1
                current = next
            i = group_num-1
            if len(group)>len(max_group_list[i]):
                while i > 0 and len(group)>len(max_group_list[i-1]):
                    max_group_list[i] = max_group_list[i-1]
                    i -= 1
                max_group_list[i] = group
            k += 1
        make_dir(dirname(filename)+'/Fix_Length_Test')
        final_lst = []
        num_lst = []
        for i in range(group_num):
            if len(max_group_list[i]) == 0:
                break
            score_lst = []
            lst = []
            num_gap_dict = {i:0 for i in range(self._min_gap,self._max_gap+1)}
            score_sum = 0
            OutFile = open('{}\Fix_Length_Test\max_group_{}.txt'.format(dirname(filename),i),'w')
            for j in max_group_list[i]:
                line = s[j][2]
                OutFile.write(line)
                matlist = list(finditer('(?<=\W)\d+(?=\W)',line))
                if len(matlist)==5:
                    start_pos = int(matlist[0].group())
                    gap = int(matlist[2].group())
                    end_pos = start_pos + gap + 43-1
                    score = int(matlist[3].group())+int(matlist[4].group())/100
                    score_sum += score
                    score_lst.append(score)
                    num_gap_dict[gap] += 1
                    lst.append((start_pos-3,end_pos))
            OutFile.close()
            length = lst[-1][1]-lst[0][0]+1
            num_lst.append(len(lst))
            final_lst.append('{:>2d}\t{:>9d}\t{:>9d}\t{:>4d}\t{:>2d}\t{}\t{:>.2f}\t{}\n'.format(i,lst[0][0],lst[-1][1],length,
                                                                                          len(score_lst),'\t'.join((str(num_gap_dict[i]) for i in range(self._min_gap,self._max_gap+1))),
                                                                                          score_sum/len(score_lst),score_lst))
        with open(dirname(filename)+'/Fix_Length_Test/result.txt','w') as OutFile:
            OutFile.write('{:>2s}\t{:>9s}\t{:>9s}\t{:>4s}\t{:>2s}\tgap{}\t{}\t{}\n'.format('#file','start_pos','end_pos','length','#pair','\tgap'.join((str(i) for i in range(self._min_gap,self._max_gap+1))),'score','score_lst'))
            self._CandidatePCRFragment.insert(tk.END,'{:>2s}\t{:>9s}\t{:>9s}\t{:>4s}\t{:>2s}\tgap{}\t{}\t{}\n'.format('#file','start_pos','end_pos','length','#pair','\tgap'.join((str(i) for i in range(self._min_gap,self._max_gap+1))),'score','score_lst'))
            for i in sorted(range(len(final_lst)),key= lambda x:num_lst[x],reverse=True):
                OutFile.write(final_lst[i])
                self._CandidatePCRFragment.insert(tk.END,final_lst[i])
        self._Choose_Cluster_Label = tk.Label(self,text='choose cluster number(seperate by blank)',font=('Calibri', 18, 'bold'))
        self._Choose_Cluster_Label.grid(row = 0, column = 3,columnspan=3)
        self._choose_cluster_number = tk.StringVar()
        self._ChooseClusterNumberEntry = tk.Entry(
            self, textvariable = self._choose_cluster_number,
            font=('Calibri', 18, 'bold'))
        self._ChooseClusterNumberEntry.grid(row = 1, column = 4,columnspan=2)
        self._cluster_finder_Button = tk.Button(self, text='Oligo Generator',font=('Calibri',18,'bold'),command=lambda :self.oligo_generator('Fix_Length_Test'))
        self._cluster_finder_Button.grid(row=3,column=2)
        return

    def oligo_generator(self,fix_test_name):
        def revcmp(dna):
            bp_dict = {'a':'t','t':'a','g':'c','c':'g'}
            s = ""
            for bp in reversed(dna):
                s += bp_dict[bp]
            return s
        self._clean()
        filenumber_list = self._choose_cluster_number.get().split()
        for i in range(len(filenumber_list)):
            target_number = i+1
            filename = '{}/{}\max_group_{}.txt'.format(dirname(self._Pair_Filename.get()),fix_test_name,filenumber_list[i])
            outfname = '{}/{}\primers_from_group_{}.txt'.format(dirname(self._Pair_Filename.get()),fix_test_name,filenumber_list[i])
            InputFile = open(filename)
            OutFile = open(outfname,'w')
            site_number = 0
            for line in InputFile:
                site_number += 1
                matlist = list(finditer('(?<=[^actg])[actg]+(?=[^actg])',line))
                left_gRNA = matlist[0].group()
                right_gRNA = matlist[1].group()
                FW_left_gRNA = 'tagg'+revcmp(left_gRNA)
                RV_left_gRNA = 'aaac'+left_gRNA
                FW_right_gRNA = 'tagg'+right_gRNA
                RV_right_gRNA = 'aaac'+revcmp(right_gRNA)
                OutFile.write('FW_F{}P{}_left\t{}\n'.format(target_number,site_number,FW_left_gRNA))
                OutFile.write('RV_F{}P{}_left\t{}\n'.format(target_number,site_number,RV_left_gRNA))
                OutFile.write('FW_F{}P{}_right\t{}\n'.format(target_number,site_number,FW_right_gRNA))
                OutFile.write('RV_F{}P{}_right\t{}\n'.format(target_number,site_number,RV_right_gRNA))
                self._CandidatePCRFragment.insert(tk.END,'FW_F{}P{}_left     {}\n'.format(target_number,site_number,FW_left_gRNA))
                self._CandidatePCRFragment.insert(tk.END,'RV_F{}P{}_left     {}\n'.format(target_number,site_number,RV_left_gRNA))
                self._CandidatePCRFragment.insert(tk.END,'FW_F{}P{}_right    {}\n'.format(target_number,site_number,FW_right_gRNA))
                self._CandidatePCRFragment.insert(tk.END,'RV_F{}P{}_right    {}\n'.format(target_number,site_number,RV_right_gRNA))
            OutFile.close()

    def _clean(self):
        self._CandidatePCRFragment.delete(0.0,tk.END)

    def _save(self):
        data = self._CandidatePCRFragment.get('0.0', tk.END)
        filename = tkfiledialog.asksaveasfilename(parent=self, title='Save as...',
                                                  filetypes=[('Text', '*.txt')])
        try:
            if not filename.endswith('.txt'):
                filename += '.txt'
            with open(filename, 'w') as w:
                w.write(data)
        except:
            tkmessagebox.showerror(title="Write File", parent = self,
                                   message = "Error: Can't open: {}".format(filename))

if __name__ == 'main':
    PCRFD_Win(None,'F:/iGEM/Software(MarkerFinder)/AI_Ora/CCN_NGG_sort_by_pos.txt').mainloop()


