from os import mkdir
from os.path import isdir
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML
from time import time
from Bio import SeqIO

def first_filter_db_lr(filename_str, out_filename_str, database_name):
    t0 = time()
    for gRNA_flag in ['CCN','NGG']:
        first_filter_db(filename_str.format(gRNA_flag), out_filename_str.format(gRNA_flag), database_name, gRNA_flag)
        print('{} first filter ends: time {}'.format(gRNA_flag, time()-t0))

def first_filter_db(in_filename, out_filename, database_name, gRNA_flag):
    t0 = time()
    blast_out_filename = first_filter_blast_db_iteration(in_filename, database_name, gRNA_flag)
    gRNA_iter = first_filter_blast_extract_iteration(in_filename,blast_out_filename,gRNA_flag)
    print('{} blast extract end: time {}'.format(in_filename,time()-t0))
    SeqIO.write(gRNA_iter, out_filename, 'fasta')

def first_filter_blast_db_iteration(in_filename,database_name,gRNA_flag):
    t0 = time()
    blast_outfile_dir = '{}_out'.format(in_filename.split('.')[0])
    if isdir(blast_outfile_dir):
        pass
    else:
        mkdir(blast_outfile_dir)
    blast_out_filename = blast_outfile_dir+"\{}_out.xml".format(gRNA_flag)
    blastn_cline = NcbiblastnCommandline(query=in_filename, db=database_name, out=blast_out_filename, evalue=10,word_size=11,reward=1,penalty=-3,gapopen=5,gapextend=2,outfmt=5,num_threads=4)
    print('{} blast start: time {}'.format(in_filename,time()-t0))
    stdout, stderr = blastn_cline()
    print('{} blast end: time {}'.format(in_filename,time()-t0))
    return blast_out_filename

def first_filter_sbjct_lr(filename_str, out_filename_str, subject_filename):
    t0 = time()
    for gRNA_flag in ['CCN','NGG']:
        first_filter_sbjct(filename_str.format(gRNA_flag), out_filename_str.format(gRNA_flag), subject_filename, gRNA_flag)
        print('{} first filter ends: time {}'.format(gRNA_flag,time()-t0))

def first_filter_sbjct(in_filename, out_filename, subject_filename,gRNA_flag):
    t0 = time()
    blast_out_filename = first_filter_blast_sbjct_iteration(in_filename, subject_filename, gRNA_flag)
    gRNA_iter = first_filter_blast_extract_iteration(in_filename,blast_out_filename,gRNA_flag,1)
    print('{} blast extract end: time {}'.format(in_filename,time()-t0))
    SeqIO.write(gRNA_iter, out_filename, 'fasta')

def first_filter_blast_sbjct_iteration(in_filename,subject_filename,gRNA_flag):
    t0 = time()
    blast_outfile_dir = '{}_self_out'.format(in_filename.split('.')[0])
    if isdir(blast_outfile_dir):
        pass
    else:
        mkdir(blast_outfile_dir)
    blast_out_filename = blast_outfile_dir+"\{}_self_out.xml".format(gRNA_flag)
    blastn_cline = NcbiblastnCommandline(query=in_filename, subject=subject_filename, out=blast_out_filename, evalue=10,word_size=11,reward=1,penalty=-3,gapopen=5,gapextend=2,outfmt=5,num_threads=4)
    print('{} blast start: time {}'.format(in_filename,time()-t0))
    stdout, stderr = blastn_cline()
    print('{} blast end: time {}'.format(in_filename,time()-t0))
    return blast_out_filename

def first_filter_blast_extract_iteration(in_filename, blast_out_filename,gRNA_flag,self_flag=0):
    def get_position(identifier):
        parts = identifier.split('|')
        try:
            assert len(parts)==2
        except:
            print(identifier)
        return parts[1]
    gRNA_dict = SeqIO.index(in_filename,'fasta',key_function=get_position)
    blast_records = NCBIXML.parse(open(blast_out_filename))
    spec_flag_dict = {position:1 for position in gRNA_dict}
    for blast_record in blast_records:
        gRNA_position = get_position(blast_record.query)
        if spec_flag_dict[gRNA_position]:
            for alignment in blast_record.alignments:
                if not spec_flag_dict[gRNA_position]: break
                for hsp in alignment.hsps:
                    if first_filter_spec_test(hsp,gRNA_flag) and (not self_flag or int(gRNA_position) != hsp.sbjct_start): # int(...start) is for self_blast test
                        spec_flag_dict[gRNA_position] = 0
                        break
    return (gRNA_dict[position] for position in gRNA_dict if spec_flag_dict[position])

def first_filter_spec_test(hsp,gRNA_flag):
    if gRNA_flag == 'CCN':
        return hsp.query_start == 1 and hsp.match.startswith('||||||||||||')

    elif gRNA_flag == 'NGG':
        return hsp.query_end == 20 and hsp.match.endswith('||||||||||||')

if __name__=='main':
    filename_str = '{}_all.fasta'
    database_name = 'db\ex_db'
    target_filename = 'ex_target.fasta'
    first_filter_out_by_db_filename_str = '{}_gRNA_first_filter_by_db.fasta'
    first_filter_out_filename_str = '{}_gRNA_first_filter.fasta'
    first_filter_db_lr(filename_str,first_filter_out_by_db_filename_str, database_name)
    first_filter_sbjct_lr(first_filter_out_by_db_filename_str,first_filter_out_filename_str, target_filename)
