from os import listdir
from os.path import isfile, dirname, basename
from Bio.Blast import NCBIXML
from Bio import SeqIO
# from itertools import cycle
from re import search
from time import time
from first_filter import first_filter_blast_db_iteration, first_filter_blast_sbjct_iteration


def record_number(records):
    k = 0
    for record in records:
        k += 1
    return k


def single_score(mismatch_lst, gRNA_flag):
    weight = [0, 0, 0.014, 0, 0, 0.395, 0.317, 0, 0.389, 0.079,
              0.445, 0.508, 0.613, 0.851, 0.732, 0.828, 0.615, 0.804, 0.685, 0.583]
    mismatch_num = len(mismatch_lst)
    if mismatch_num == 0:
        return 100
    if mismatch_num == 1:
        return 100*(1-weight[20-mismatch_lst[0]]) if gRNA_flag == 'CCN' else 100*(1-weight[mismatch_lst[0]-1])
    distance_sum = 0
    for i in range(1, mismatch_num+1):
        distance_sum += mismatch_lst[i-1]*(2*i-mismatch_num-1)
    mean_distance = 2.0*distance_sum/mismatch_num/(mismatch_num-1)-1
    w = 1
    if gRNA_flag == 'CCN':
        for i in mismatch_lst:
            w *= 1-weight[20-i]
    else:
        for i in mismatch_lst:
            w *= 1-weight[i-1]
    singlescore = 100*w / ((19-mean_distance)/19.0*4+1.0)/(mismatch_num**2)
    return singlescore


def before_query_analysis(hsp, target_sequence, subject_sequence, mismatch_list, rev=False):
    subject_len = len(subject_sequence)
    for i in range(1, hsp.query_start):
        if rev:
            pos = hsp.sbjct_start + hsp.query_start-i
            if pos > subject_len:
                pos -= subject_len
        else:
            pos = hsp.sbjct_start - hsp.query_start+i
            if pos <= 0:
                pos += subject_len
        if subject_sequence[pos-1] != target_sequence[i-1]:
            mismatch_list.append(i)
            # print(i,pos,Escherichia_coli_genome[pos-1],sequence[i-1])


def align_analysis(hsp, mismatch_list, gap_list):
    gapnumber = 0
    i = hsp.query_start
    for aa in hsp.query:
        if aa == '-':
            gap_list.append(i-gapnumber)
            gapnumber += 1
        elif aa.lower() != hsp.sbjct[i-hsp.query_start].lower():
            mismatch_list.append(i-gapnumber)
        i += 1


def after_query_analysis(hsp, target_sequence, subject_sequence, mismatch_list, rev=False):
    subject_len = len(subject_sequence)
    for i in range(hsp.query_end+1, 21):
        if rev:
            pos = hsp.sbjct_end + hsp.query_end - i
            if pos <= 0:
                pos += subject_len
        else:
            pos = hsp.sbjct_end - hsp.query_end + i
            if pos > subject_len:
                pos -= subject_len
        if subject_sequence[pos-1] != target_sequence[i-1]:
            mismatch_list.append(i)


def score_blast_extract(score_gRNA_dict, maxscore_gRNA_dict, choose_gRNA_dict, subject_filename, blast_outfile_dir, gRNA_flag):
    for filename in listdir(blast_outfile_dir):
        blast_out_filename = blast_outfile_dir + '\\'+filename
        result_handle = open(blast_out_filename)
        subject_record = SeqIO.read(subject_filename, 'fasta')
        subject_sequence = subject_record.seq
        blast_records = NCBIXML.parse(result_handle)
        filter_func = lambda rec: rec.query.split('|')[1] in choose_gRNA_dict
        for blast_record in filter(filter_func, blast_records):
            gRNA_position = blast_record.query.split('|')[1]
            target_sequence = choose_gRNA_dict[gRNA_position]
            for alignment in blast_record.alignments:
                for hsp in alignment.hsps:
                    mismatch_list = []
                    gap_list = []
                    if hsp.sbjct_start < hsp.sbjct_end:
                        before_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list)
                        align_analysis(hsp, mismatch_list, gap_list)
                        after_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list)
                    else:
                        before_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list, rev=True)
                        align_analysis(hsp, mismatch_list, gap_list)
                        after_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list, rev=True)
                    new_score = single_score(mismatch_list, gRNA_flag)
                    if new_score > maxscore_gRNA_dict[gRNA_position] and new_score != 100:
                        maxscore_gRNA_dict[gRNA_position] = new_score
                    score_gRNA_dict[gRNA_position] += new_score


def score_blast_db_extract(score_gRNA_dict, maxscore_gRNA_dict, choose_gRNA_dict, subject_filename, blast_outfile_dir, gRNA_flag):
    for filename in listdir(blast_outfile_dir):
        blast_out_filename = blast_outfile_dir + '\\'+filename
        result_handle = open(blast_out_filename)
        subject_record_dict = SeqIO.index(subject_filename, 'fasta')
        filedir = dirname(blast_outfile_dir)
        with open('{}\index.txt'.format(filedir), 'w') as outfile:
            for key in subject_record_dict:
                outfile.write('{}\n'.format(key))
        blast_records = NCBIXML.parse(result_handle)
        # filter_func = lambda rec: rec.query.split('|')[1] in choose_gRNA_dict
        # k = 0
        # for blast_record in filter(filter_func,blast_records):
        for blast_record in blast_records:
            gRNA_position = blast_record.query.split('|')[1]
            target_sequence = choose_gRNA_dict[gRNA_position]
            for alignment in blast_record.alignments:
                subject_sequence = subject_record_dict[
                    alignment.hit_def.split()[0]]
                for hsp in alignment.hsps:
                    mismatch_list = []
                    gap_list = []
                    if hsp.sbjct_start < hsp.sbjct_end:
                        before_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list)
                        align_analysis(hsp, mismatch_list, gap_list)
                        after_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list)
                    else:
                        before_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list, rev=True)
                        align_analysis(hsp, mismatch_list, gap_list)
                        after_query_analysis(
                            hsp, target_sequence, subject_sequence, mismatch_list, rev=True)
                    new_score = single_score(mismatch_list, gRNA_flag)
                    if new_score > maxscore_gRNA_dict[gRNA_position] and new_score != 100:
                        maxscore_gRNA_dict[gRNA_position] = new_score
                    score_gRNA_dict[gRNA_position] += new_score


def second_filter_prepare(filename_str, target_filename, database_name):
    for gRNA_flag in ['CCN', 'NGG']:
        in_filename = filename_str.format(gRNA_flag)
        first_filter_blast_db_iteration(in_filename, database_name, gRNA_flag)
        first_filter_blast_sbjct_iteration(
            in_filename, target_filename, gRNA_flag)


def gRNA_score(choose_gRNA_filename, target_filename, target_blast_outfile_dir, subject_filename, subject_blast_outfile_dir, out_filename, gRNA_flag):
    def get_position(identifier):
        parts = identifier.split('|')
        try:
            assert len(parts) == 2
        except:
            print(identifier)
        return parts[1]
    t0 = time()
    choose_gRNA_dict = SeqIO.index(
        choose_gRNA_filename, 'fasta', key_function=get_position)
    score_gRNA_dict = {gRNA_position: 0 for gRNA_position in choose_gRNA_dict}
    maxscore_gRNA_dict = {
        gRNA_position: 0 for gRNA_position in choose_gRNA_dict}
    print('{} blast out extract start: time {}'.format(gRNA_flag, time()-t0))
    filedir = dirname(choose_gRNA_filename)
    if not isfile('{}\pre_{}'.format(filedir, basename(out_filename))):
        score_blast_db_extract(score_gRNA_dict, maxscore_gRNA_dict,
                               choose_gRNA_dict, subject_filename, subject_blast_outfile_dir, gRNA_flag)
        print('{} blast oral out extract end: time {}'.format(
            gRNA_flag, time()-t0))
        with open('{}\pre_{}'.format(filedir, basename(out_filename)), 'w') as OutFile:
            for gRNA_position in sorted(score_gRNA_dict, key=lambda x: int(x)):
                OutFile.write('{}\t{}\t{}\t{}\n'.format(choose_gRNA_dict[
                              gRNA_position].seq, gRNA_position, score_gRNA_dict[gRNA_position], maxscore_gRNA_dict[gRNA_position]))
        print('pre {} save end: time {}'.format(gRNA_flag, time()-t0))
    else:
        print('pre {} save end: time {}'.format(gRNA_flag, time()-t0))
        Data = open('{}\pre_{}'.format(filedir, basename(out_filename)))
        for line in Data:
            a = line.strip().split('\t')
            gRNA_position = a[1]
            score = a[2]
            maxscore = a[3]
            score_gRNA_dict.setdefault(gRNA_position, score)
            maxscore_gRNA_dict.setdefault(gRNA_position, maxscore)
    score_blast_extract(score_gRNA_dict, maxscore_gRNA_dict,
                        choose_gRNA_dict, target_filename, target_blast_outfile_dir, gRNA_flag)
    print('{} blast self out extract end: time {}'.format(
        gRNA_flag, time()-t0))
    with open(out_filename, 'w') as OutFile:
        for gRNA_position in sorted(score_gRNA_dict, key=lambda x: int(x)):
            try:
                OutFile.write('{}\t{}\t{}\t{}\n'.format(choose_gRNA_dict[
                              gRNA_position].seq, gRNA_position, 10000/score_gRNA_dict[gRNA_position], maxscore_gRNA_dict[gRNA_position]))
            except ZeroDivisionError:
                OutFile.write('{}\t{}\t{}\t{}\n'.format(choose_gRNA_dict[
                              gRNA_position].seq, gRNA_position, score_gRNA_dict[gRNA_position], maxscore_gRNA_dict[gRNA_position]))
    print('{} save end: time {}'.format(gRNA_flag, time()-t0))
    # return choose_gRNA_dict


def gRNA_score_filter(infname, outfname, gRNA_score_threshold=45, max_score_threshold=2):
    InputFile = open(infname)
    lst = []
    regexp = r'a{7,}|t{7,}|g{7,}|c{7,}'
    for line in InputFile:
        a = line.strip()
        if search(regexp, a) is None:
            a = a.split()
            score = float(a[2])
            max_score = float(a[3])
            if score > gRNA_score_threshold and max_score <= max_score_threshold:
                lst.append(line)
    InputFile.close()
    with open(outfname, 'w') as OutFile:
        for line in lst:
            OutFile.write(line)


def second_filter(choose_gRNA_filename, target_filename, target_blast_outfile_dir, subject_filename, subject_blast_outfile_dir, gRNA_flag, gRNA_score_threshold=45, max_score_threshold=2):
    filedir = dirname(choose_gRNA_filename)
    out_filename = '{}\score_{}_gRNA.txt'.format(filedir, gRNA_flag)
    gRNA_score(choose_gRNA_filename, target_filename, target_blast_outfile_dir,
               subject_filename, subject_blast_outfile_dir, out_filename, gRNA_flag)
    choose_filename = '{}\choose_{}_gRNA.txt'.format(filedir, gRNA_flag)
    gRNA_score_filter(
        out_filename, choose_filename, gRNA_score_threshold, max_score_threshold)


def second_filter_lr(first_filter_out_filename_str, target_filename, subject_filename, database_name):
    t0 = time()
    second_filter_prepare(
        first_filter_out_filename_str, target_filename, database_name)
    target_blast_outfile_dir_str = '{}_self_out'.format(
        first_filter_out_filename_str.split('.')[0])
    subject_blast_outfile_dir_str = '{}_out'.format(
        first_filter_out_filename_str.split('.')[0])
    for gRNA_flag in ['CCN', 'NGG']:
        target_blast_outfile_dir = target_blast_outfile_dir_str.format(
            gRNA_flag)
        subject_blast_outfile_dir = subject_blast_outfile_dir_str.format(
            gRNA_flag)
        choose_gRNA_filename = first_filter_out_filename_str.format(gRNA_flag)
        second_filter(choose_gRNA_filename, target_filename, target_blast_outfile_dir,
                      subject_filename, subject_blast_outfile_dir, gRNA_flag)
        print('{} second filter ends: time {}'.format(gRNA_flag, time()-t0))


if __name__ == 'main':
    database_name = 'db\ex_db'
    target_filename = 'ex_target.fasta'
    filename_str = '{}_gRNA_first_filter.fasta'
    first_filter_out_filename_str = '{}_gRNA_first_filter.fasta'
    for gRNA_flag in ['CCN','NGG']:
        in_filename = filename_str.format(gRNA_flag)
        first_filter_blast_db_iteration(in_filename,database_name,gRNA_flag)
        first_filter_blast_sbjct_iteration(in_filename,target_filename,gRNA_flag)
    subject_filename = 'ex_db.fasta'
    subject_blast_outfile_dir_str = '{}_gRNA_first_filter_out'
    target_blast_outfile_dir_str = '{}_gRNA_first_filter_by_db_out'
    second_filter_lr(first_filter_out_filename_str,target_filename,subject_filename,database_name)
